/******************************************************************************
 * File: bwagent.cpp
 *
 * The main file for the Black & White agent.
 *
 * @author      Samuel Audet
 * @version     1.0, 2006-01-30
 */
#include <unistd.h>

#include <iostream>
#include <fstream>
using namespace std;

#include "Node.h"
#include "TransTable.h"

/**
 * Train the program by making it play games against itself
 * with automatically varying weights on features.
 * Will log the results in training_log.txt.
 *
 * Does not really work...
 */
void train() {
    ofstream log("training_log.txt");
    Node *theGame = Node::theStack.pop();

    for (Node::PROTECTED_WEIGHT[0] = 1; Node::PROTECTED_WEIGHT[0] < 10; Node::PROTECTED_WEIGHT[0]+=2)
    for (Node::DOUBLY_PROTECTED_WEIGHT[0] = 1; Node::DOUBLY_PROTECTED_WEIGHT[0] < 10; Node::DOUBLY_PROTECTED_WEIGHT[0]+=2)
    for (Node::LINEUP_WEIGHT[0] = 1; Node::LINEUP_WEIGHT[0] < 10; Node::LINEUP_WEIGHT[0]+=2) 
    for (Node::ATTACK_WEIGHT[0] = 1; Node::ATTACK_WEIGHT[0] < 10; Node::ATTACK_WEIGHT[0]+=2) {    

        cout << Node::PROTECTED_WEIGHT[0] << " " << Node::DOUBLY_PROTECTED_WEIGHT[0] 
             << " " << Node::LINEUP_WEIGHT[0] << " " << Node::ATTACK_WEIGHT[0] << " against "
             << Node::PROTECTED_WEIGHT[1] << " " << Node::DOUBLY_PROTECTED_WEIGHT[1] 
             << " " << Node::LINEUP_WEIGHT[1] << " " << Node::ATTACK_WEIGHT[1] << endl;

        log  << Node::PROTECTED_WEIGHT[0] << " " << Node::DOUBLY_PROTECTED_WEIGHT[0] 
             << " " << Node::LINEUP_WEIGHT[0] << " " << Node::ATTACK_WEIGHT[0] << " against "
             << Node::PROTECTED_WEIGHT[1] << " " << Node::DOUBLY_PROTECTED_WEIGHT[1] 
             << " " << Node::LINEUP_WEIGHT[1] << " " << Node::ATTACK_WEIGHT[1];
        log.flush();

        theGame->clear(PLAYER0);
        int turn = 0;
        while (true) {
            turn++;

            cout << "Turn " << turn << ": " << theGame->getPlayer() << "'s move: ";
            int value = theGame->play(-1, true, true);
            cout << "Value: " << value << endl;
            theGame->print();

            // now the other player's turn.. did he lose?
            if (theGame->hasLost())
                break; 

            cout << "Turn " << turn << ": " << theGame->getPlayer() << "'s move: ";
            value = theGame->play(-1, true, true);
            cout << "Value: " << value << endl;
            theGame->print();

            // now the other player's turn.. did he lose?
            if (theGame->hasLost())
                break;          

            if (turn > 50) { 
                cout << " Draw" << endl;
                log  << " Draw" << endl;
                log.flush();
                break; 
            }
        }

        if (theGame->hasLost()) {
            if (theGame->getPlayer() == PLAYER1) {
                // if it's PLAYER1's turn, he lost, so PLAYER0 won
                // and since we are modifying PLAYER0's parameter,
                // it means these new parameters should be better
                cout << "  New parameters won!" << endl;
                log  << "  New parameters won!" << endl;
                log.flush();

                Node::PROTECTED_WEIGHT[1]        = Node::PROTECTED_WEIGHT[0];
                Node::DOUBLY_PROTECTED_WEIGHT[1] = Node::DOUBLY_PROTECTED_WEIGHT[0];
                Node::LINEUP_WEIGHT[1]           = Node::LINEUP_WEIGHT[0];
                Node::ATTACK_WEIGHT[1]           = Node::ATTACK_WEIGHT[0];
            } else {
                cout << "  New parameters lost." << endl;
                log  << "  New parameters lost." << endl;
                log.flush();    
            }
        }
    }
}

int main(int argc, char *argv[]) {
    char c;
    bool doTraining = false;
    bool computerIsWhite = false;
    bool readRandomFile = false;
    bool wantHelp = false;
    while((c = getopt(argc, argv, "twrh")) != -1) {
        switch(c) {
        case 't': doTraining = true; break;
        case 'w': computerIsWhite = true; break;
        case 'r': readRandomFile = true; break;
        case 'h': 
        default:  wantHelp = true; break;
            
        }
    }

    // print help if requested and exit
    if (wantHelp) {
        cerr << "-t : run in training mode" << endl 
             << "-r : read random numbers from random_numbers.txt instead of generating them" << endl
             << "-w : computer starts with white" << endl;
        return 1;
    }

    Node::initCodes(readRandomFile);
    cout << "Time per turn: " << TIME_PER_TURN << " seconds." << endl;

    if (doTraining) {
        train();
    } else {
        // start with clear board
        Node *theGame = Node::theStack.pop();  
        theGame->clear(PLAYER1); 

        int turn = 1;
        if (computerIsWhite) {
            cout << "Turn " << ++turn/2 << ": " << theGame->getPlayer() << "'s move (computer): ";
            int value = theGame->play(-1, true, true);
            cout << "Value: " << value << endl;
            theGame->print();
        }

        while (true) {
            // human's move
            cout << "Turn " << ++turn/2 << ": " << theGame->getPlayer() << "'s move (you): ";
            // parse and wait a valid move from human...
            while (!theGame->parse()) { }
            theGame->print();
            
            // turn switches to computer ... did it lose?
            if (theGame->hasLost()) {
                cout << "You win!" << endl;         
                break;
            }

            // computer's move
            cout << "Turn " << ++turn/2 << ": " << theGame->getPlayer() << "'s move (computer): ";
            int value = theGame->play(-1, true, true);
            cout << "Value: " << value << endl;
            theGame->print();

            // now the human player's turn.. did he lose?
            if (theGame->hasLost()) {
                cout << "You lose." << endl;
                break;
            }
        }
    }

    return 0;
}
